use regex::Regex;

use std::str::FromStr;

#[derive(Clone, Debug, Serialize, Deserialize, PartialEq, Eq)]
pub enum OsFamily {
    Debian,
    Archlinux,
}

impl OsFamily {
    pub fn as_str(&self) -> &'static str {
        match self {
            Self::Debian => "debian",
            Self::Archlinux => "archlinux",
        }
    }
}

#[derive(Clone, Debug, Serialize, Deserialize, PartialEq, Eq)]
pub struct OsFact {
    family: OsFamily,
    version_int: Option<usize>,
    version_str: Option<String>,
}

impl OsFact {
    /// Returns an [`OsFact`] from the standard `/etc/os-release` file
    pub fn new() -> OsFact {
        let content = std::fs::read_to_string("/etc/os-release").expect("Failed to read `/etc/os-release file.`");
        OsFact::from_str(&content)
    }

    /// Returns an [`OsFact`] from the passed string slice (used for tests)
    pub fn from_str(os_release: &str) -> OsFact {
        let re = Regex::new(r#"(?m)^ID=(.*)"#).unwrap();
        if let Some(captures) = re.captures(&os_release) {
            if let Some(id) = captures.get(1) {
                match id.as_str() {
                    "debian" => {
                        let re = Regex::new(r#"(?m)^VERSION="(\d+) \((.*)\)"$"#).unwrap();
                        if let Some(captures) = re.captures(&os_release) {
                            let (v_int, v_str) = (captures.get(1), captures.get(2));
                            match (v_int, v_str) {
                                (Some(v_int), Some(v_str)) => {
                                    let v_int = usize::from_str(v_int.as_str()).expect(
                                        &format!("The OS version from /etc/os-release {} could not be parsed as integer.", v_int.as_str())
                                    );
                                    let v_str = v_str.as_str().to_string();
                                    OsFact {
                                        family: OsFamily::Debian,
                                        version_int: Some(v_int),
                                        version_str: Some(v_str),
                                    }
                                }, _ => {
                                    panic!("OS was detected as Debian but some versioning info is missing. Integer: {:?}, String: {:?}", v_int, v_str);
                                }
                                
                            }
                            
                        } else {
                            panic!("OS was detected as Debian but version information could not be acquired.");
                        }
                    }, "arch" => OsFact {
                        family: OsFamily::Archlinux,
                        version_int: None,
                        version_str: None,
                    }, _ => {
                        panic!("OS Family not recognized: {}", id.as_str());
                    }
                }
            } else {
                panic!("regex capture `ID` does not exist in /etc/os-release");
            }
        } else {
            panic!("/etc/os-release does not contain `ID` line");
        }
    }

    pub fn family(&self) -> &OsFamily {
        &self.family
    }
}

#[cfg(test)]
mod tests {
    use super::{OsFact, OsFamily};
    
    #[test]
    fn can_parse_debian_bookworm() {
        let os_release = r#"
PRETTY_NAME="Debian GNU/Linux 12 (bookworm)"
NAME="Debian GNU/Linux"
VERSION_ID="12"
VERSION="12 (bookworm)"
VERSION_CODENAME=bookworm
ID=debian
HOME_URL="https://www.debian.org/"
SUPPORT_URL="https://www.debian.org/support"
BUG_REPORT_URL="https://bugs.debian.org/"
"#;
        let fact = OsFact::from_str(os_release);
        assert_eq!(fact, OsFact {
            family: OsFamily::Debian,
            version_int: Some(12),
            version_str: Some("bookworm".to_string())
        });
    }


    #[test]
    fn can_parse_debian_bullseye() {
        let os_release = r#"
PRETTY_NAME="Debian GNU/Linux 11 (bullseye)"
NAME="Debian GNU/Linux"
VERSION_ID="11"
VERSION="11 (bullseye)"
VERSION_CODENAME=bullseye
ID=debian
HOME_URL="https://www.debian.org/"
SUPPORT_URL="https://www.debian.org/support"
BUG_REPORT_URL="https://bugs.debian.org/"
"#;
        let fact = OsFact::from_str(os_release);
        assert_eq!(fact, OsFact {
            family: OsFamily::Debian,
            version_int: Some(11),
            version_str: Some("bullseye".to_string())
        });
    }

    #[test]
    fn can_parse_archlinux() {
        let os_release = r#"
NAME="Arch Linux"
PRETTY_NAME="Arch Linux"
ID=arch
BUILD_ID=rolling
ANSI_COLOR="38;2;23;147;209"
HOME_URL="https://archlinux.org/"
DOCUMENTATION_URL="https://wiki.archlinux.org/"
SUPPORT_URL="https://bbs.archlinux.org/"
BUG_REPORT_URL="https://bugs.archlinux.org/"
PRIVACY_POLICY_URL="https://terms.archlinux.org/docs/privacy-policy/"
LOGO=archlinux-logo
"#;
        let fact = OsFact::from_str(os_release);
        assert_eq!(fact, OsFact {
            family: OsFamily::Archlinux,
            version_int: None,
            version_str: None
        });
    }

}