use std::path::{Path, PathBuf};

use crate::modules::Module;
use super::{CommandModule, CommandError, CommandStatus};
use super::condition::{SomeCondition};

use crate::utils::cmd::Cmd;

pub struct NoCmd;

pub struct NoCondition;
#[derive(Clone, Debug, Serialize)]
pub struct CreatesCondition(pub PathBuf);
#[derive(Clone, Debug, Serialize)]
pub struct RemovesCondition(pub PathBuf);
#[derive(Clone, Debug, Serialize)]
pub struct BothConditions(pub CreatesCondition, pub RemovesCondition);

/// Build a [`CommandModule`].
/// You need to define the program/args before you can set the directory,
/// or populate stdin.
pub struct CommandBuilder<Program, Condition> {
    cmd: Program,
    condition: Option<Condition>,
}

impl CommandBuilder<NoCmd, NoCondition> {
    pub fn new() -> Self {
        CommandBuilder {
            cmd: NoCmd,
            condition: None,
        }
    }
}

impl<Condition> CommandBuilder<NoCmd, Condition> {
    pub fn program<T: AsRef<str>>(self, program: T) -> CommandBuilder<Cmd, Condition> {
        let Self { condition, .. } = self;

        CommandBuilder {
            cmd: Cmd::new(program),
            condition,
        }
    }
}

impl<Condition> CommandBuilder<Cmd, Condition> {
    pub fn arg<T: AsRef<str>>(self, arg: T)  -> CommandBuilder<Cmd, Condition> {
        let Self { cmd, condition, .. } = self;

        Self {
            cmd: cmd.arg(arg),
            condition,
        }
    }

    pub fn args<T: AsRef<str>>(self, args: & [ T ])  -> CommandBuilder<Cmd, Condition> {
        let Self { cmd, condition, .. } = self;

        Self {
            cmd: cmd.args(args),
            condition,
        }
    }

    pub fn stdin<T: AsRef<str>>(self, input: T)  -> CommandBuilder<Cmd, Condition> {
        let Self { cmd, condition, .. } = self;

        CommandBuilder {
            cmd: cmd.stdin(input),
            condition,
        }
    }

    pub fn chdir<T: AsRef<str>>(self, dir: T)  -> CommandBuilder<Cmd, Condition> {
        let Self { cmd, condition, .. } = self;
        
        CommandBuilder {
            cmd: cmd.chdir(dir),
            condition,
        }
    }
}

impl<Args> CommandBuilder<Args, NoCondition> {
    pub fn creates<T: AsRef<Path>>(self, path: T) -> CommandBuilder<Args, CreatesCondition> {
        let Self { cmd, .. } = self;

        CommandBuilder {
            cmd,
            condition: Some(CreatesCondition(path.as_ref().to_path_buf())),
        }
    }

    pub fn removes<T: AsRef<Path>>(self, path: T) -> CommandBuilder<Args, RemovesCondition> {
        let Self { cmd, .. } = self;

        CommandBuilder {
            cmd,
            condition: Some(RemovesCondition(path.as_ref().to_path_buf())),
        }
    }
}

impl<Args> CommandBuilder<Args, CreatesCondition> {
    pub fn removes<T: AsRef<Path>>(self, path: T) -> CommandBuilder<Args, BothConditions> {
        let Self {cmd, condition, .. } = self;

        CommandBuilder {
            cmd,
            condition: Some(BothConditions(condition.unwrap(), RemovesCondition(path.as_ref().to_path_buf()))),
        }
    }
}

impl<Args> CommandBuilder<Args, RemovesCondition> {
    pub fn creates<T: AsRef<Path>>(self, path: T) -> CommandBuilder<Args, BothConditions> {
        let Self { cmd, condition, .. } = self;

        CommandBuilder {
            cmd,
            condition: Some(BothConditions(CreatesCondition(path.as_ref().to_path_buf()), condition.unwrap())),
        }
    }
}

impl<Condition: Into<Option<SomeCondition>>> CommandBuilder<Cmd, Condition> {
    pub fn build(self) -> CommandModule {
        let Self { cmd, condition, .. } = self;       

        let condition: Option<SomeCondition> = condition.map(|x| x.into()).flatten();
        
        CommandModule {
            cmd,
            condition,
        }
    }

    pub fn run(self) -> Result<CommandStatus, CommandError> {
        CommandModule::run(self.build()).into()
    }
}
