use crate::utils::cmd::{Cmd, CmdOutput};

use super::{PackageError, PackageList, PackageManager};

/// This structure is mostly used as an indirection for testing purposes.
/// But you can use it to build your own pacman commands.
#[derive(Clone, Debug, PartialEq, Eq, Serialize)]
pub struct PacmanCmd(pub Cmd);

impl PacmanCmd {
    pub fn new() -> PacmanCmd {
        PacmanCmd(
            Cmd::new("pacman").arg("--noconfirm").arg("--needed")
        )
    }

    pub fn update() -> PacmanCmd {
        let mut cmd = Self::new();
        cmd.0 = cmd.0.arg("-Sy");
        cmd
    }

    pub fn install(list: PackageList) -> PacmanCmd {
        let mut cmd = Self::new();
        cmd.0 = cmd.0.arg("-S").args(list.list());
        cmd
    }

    pub fn remove(list: PackageList) -> PacmanCmd {
        let mut cmd = Self::new();
        cmd.0 = cmd.0.arg("-R").args(list.list());
        cmd
    }

    pub fn run(self) -> Result<CmdOutput, std::io::Error> {
        self.0.run()
    }
}


#[derive(Debug)]
pub struct PacmanManager;

impl PackageManager for PacmanManager {
    fn name(&self) -> &'static str {
        "pacman"
    }
    
    fn update(&self) -> Result<(), PackageError> {
        let res = PacmanCmd::update().run()?;

        if ! res.success {
            return Err(PackageError::CmdFail(res));
        }

        Ok(())
    }

    fn install(&self, list: PackageList) -> Result<(), PackageError> {
        let res = PacmanCmd::install(list).run()?;

        if ! res.success {
            return Err(PackageError::CmdFail(res));
        }

        Ok(())
    }

    fn remove(&self, list: PackageList) -> Result<(), PackageError> {
        let res = PacmanCmd::remove(list).run()?;

        if ! res.success {
            return Err(PackageError::CmdFail(res));
        }

        Ok(())
    }
}

#[cfg(test)]
mod tests {
    use super::*;
    use super::super::IntoPackageList;
    use std::collections::HashMap;

    #[test]
    pub fn update_cmd() {
        let cmd = PacmanCmd::update();
        let mut args: Vec<&str> = vec!("--noconfirm", "--needed");
        args.push("-Sy");

        assert_eq!(
            cmd.0,
            Cmd {
                program: "pacman".to_string(),
                args: args.iter().map(|x| x.to_string()).collect(),
                stdin: None,
                chdir: None,
                env: HashMap::new(),
            }
        )
    }

    #[test]
    pub fn install_cmd() {
        let cmd = PacmanCmd::install(vec!("sl").into_package_list());
        let mut args: Vec<&str> = vec!("--noconfirm", "--needed");
        args.extend(& [ "-S", "sl" ]);

        assert_eq!(
            cmd.0,
            Cmd {
                program: "pacman".to_string(),
                args: args.iter().map(|x| x.to_string()).collect(),
                stdin: None,
                chdir: None,
                env: HashMap::new(),
            }
        )
    }

    #[test]
    pub fn install_cmd_multi() {
        let cmd = PacmanCmd::install(vec!("sl", "hello").into_package_list());
        let mut args: Vec<&str> = vec!("--noconfirm", "--needed");
        args.extend(& [ "-S", "sl", "hello" ]);

        assert_eq!(
            cmd.0,
            Cmd {
                program: "pacman".to_string(),
                args: args.iter().map(|x| x.to_string()).collect(),
                stdin: None,
                chdir: None,
                env: HashMap::new(),
            }
        )
    }

    #[test]
    pub fn remove_cmd() {
        let cmd = PacmanCmd::remove(vec!("sl").into_package_list());
        let mut args: Vec<&str> = vec!("--noconfirm", "--needed");
        args.extend(& [ "-R", "sl" ]);

        assert_eq!(
            cmd.0,
            Cmd {
                program: "pacman".to_string(),
                args: args.iter().map(|x| x.to_string()).collect(),
                stdin: None,
                chdir: None,
                env: HashMap::new(),
            }
        )
    }

    #[test]
    pub fn remove_cmd_multi() {
        let cmd = PacmanCmd::remove(vec!("sl", "hello").into_package_list());
        let mut args: Vec<&str> = vec!("--noconfirm", "--needed");
        args.extend(& [ "-R", "sl", "hello" ]);

        assert_eq!(
            cmd.0,
            Cmd {
                program: "pacman".to_string(),
                args: args.iter().map(|x| x.to_string()).collect(),
                stdin: None,
                chdir: None,
                env: HashMap::new(),
            }
        )
    }
}
